import logging
import pytest
import random
import socket

from kesslerav.protocol2k.io import \
  Command, Instruction, Codec, TcpDevice, TcpEndpoint, _VALID_RANGE

from fakes import FakeSocket

class TestCommand:
  def test_is_supported_returns_true_for_known_command_id(self):
    valid_cmd = gen_valid_cmd()

    result = Command.is_supported(valid_cmd)

    assert result is True

  def test_is_supported_returns_false_for_unknown_command_id(self):
    invalid_cmd_id = gen_invalid_cmd_id()

    result = Command.is_supported(invalid_cmd_id)

    assert result is False

class TestInstruction:
  def test_name_returns_command_name_for_supported_commands(self):
    valid_cmd = gen_valid_cmd()

    result = Instruction(valid_cmd)

    assert result.name == valid_cmd.name

  def test_name_returns_unsupported_for_unsupported_commands(self):
    invalid_cmd_id = gen_invalid_cmd_id()

    result = Instruction(invalid_cmd_id)

    assert result.name == Instruction.UNSUPPORTED_COMMAND_NAME

  def test_id_returns_command_id_for_supported_commands(self):
    valid_cmd = gen_valid_cmd()

    result = Instruction(valid_cmd)

    assert result.id == valid_cmd.value

  def test_id_returns_unsupported_command_id_for_unsupported_commands(self):
    invalid_cmd_id = gen_invalid_cmd_id()

    result = Instruction(invalid_cmd_id)

    assert result.id == invalid_cmd_id

  def test_input_value_returns_value_when_valid(self):
    valid_cmd = gen_valid_cmd()
    input_value = gen_valid_io_value()

    result = Instruction(valid_cmd, input_value)

    assert result.input_value == input_value

  def test_input_value_raises_exception_when_invalid(self):
    valid_cmd = gen_valid_cmd()
    input_value = gen_invalid_io_value()

    with pytest.raises(ValueError):
      Instruction(valid_cmd, input_value)

  def test_output_value_returns_value_when_valid(self):
    valid_cmd = gen_valid_cmd()
    input_value = gen_valid_io_value()
    output_value = gen_valid_io_value()

    result = Instruction(valid_cmd, input_value, output_value)

    assert result.output_value == output_value
  
  def test_output_value_raises_exception_when_invalid(self):
    valid_cmd = gen_valid_cmd()
    input_value = gen_valid_io_value()
    output_value = gen_invalid_io_value()

    with pytest.raises(ValueError):
      Instruction(valid_cmd, input_value, output_value)

  def test_machine_id_returns_default_when_not_provided(self):
    valid_cmd = gen_valid_cmd()

    result = Instruction(valid_cmd)

    assert result.machine_id == Instruction.DEFAULT_MACHINE_ID

  def test_machine_id_returns_value_when_valid(self):
    valid_cmd = gen_valid_cmd()
    input_value = gen_valid_io_value()
    output_value = gen_valid_io_value()
    machine_id = random.randrange(1, 10)

    result = Instruction(valid_cmd, input_value, output_value, machine_id)

    assert result.machine_id == machine_id

  def test_machine_id_raises_exception_when_invalid(self):
    valid_cmd = gen_valid_cmd()
    input_value = gen_valid_io_value()
    output_value = gen_valid_io_value()
    machine_id = gen_invalid_io_value()

    with pytest.raises(ValueError):
      Instruction(valid_cmd, input_value, output_value, machine_id)

  def test_frame_returns_list_of_data(self):
    valid_cmd = gen_valid_cmd()
    input_value = gen_valid_io_value()
    output_value = gen_valid_io_value()
    machine_id = gen_valid_io_value()
    expected = [valid_cmd, input_value, output_value, machine_id]
    instruction = Instruction(*expected)

    result = instruction.frame

    assert result == expected

class TestCodec:
  def test_encode_generates_valid_bytes(self):
    expected = b'\x1f\x80\x80\xc1'
    cmd = Instruction(Command.QUERY_PANEL_LOCK)

    result = Codec.encode(cmd)

    assert result == expected

  def test_decode_hydrates_from_request_bytes(self):
    req_bytes = b'\x1f\x80\x80\xc1'
    expected = Instruction(Command.QUERY_PANEL_LOCK)

    result = Codec.decode(req_bytes)

    assert result == expected

  def test_decode_hydrates_from_response_bytes(self):
    resp_bytes = b'\x5f\x80\x80\xc1'
    expected = Instruction(Command.QUERY_PANEL_LOCK)

    result = Codec.decode(resp_bytes)

    assert result == expected

  def test_decode_encode_returns_starting_value(self):
    input = b'\x1f\x80\x80\xc1'
    cmd = Codec.decode(input)

    result = Codec.encode(cmd)

    assert result == input

  def test_encode_decode_returns_starting_value(self):
    input = Instruction(Command.QUERY_PANEL_LOCK)
    req_bytes = Codec.encode(input)

    result = Codec.decode(req_bytes)

    assert result == input

class TestTcpEndpoint:
  _host = 'localhost'

  def test_host_returns_specified_value(self):
    expected = self._host

    sut = TcpEndpoint(expected)

    assert sut.host == expected

  def test_port_returns_specified_value_when_provided(self):
    expected = gen_port()

    sut = TcpEndpoint(self._host, expected)

    assert sut.port == expected

  def test_port_returns_default_value_when_not_provided(self):
    port = None

    sut = TcpEndpoint(self._host, port)

    assert sut.port == TcpEndpoint.DEFAULT_PORT

  def test_timeout_returns_specified_value_when_provided(self):
    expected = 0.5

    sut = TcpEndpoint(self._host, gen_port(), expected)

    assert sut.timeout_sec == expected

  def test_timeout_returns_none_when_none_provided(self):
    sut = TcpEndpoint(self._host, gen_port(), None)

    assert sut.timeout_sec is None

  def test_timeout_returns_default_when_elided(self):
    sut = TcpEndpoint(self._host, gen_port())

    assert sut.timeout_sec == TcpEndpoint.DEFAULT_TIMEOUT_SEC

class TestTcpDevice:
  def test_creates_connection_using_specified_endpoint_details(self, monkeypatch: pytest.MonkeyPatch):
    expected_host = '10.0.0.1'
    expected_port = 1337
    expected_timeout = 0.101
    captured_host = False
    captured_port = False
    captured_timeout_sec = False
    fake_socket = FakeSocket()
    def capture_create_connection(host_port_tuple, timeout_sec):
      nonlocal captured_host, captured_port, captured_timeout_sec
      captured_host = host_port_tuple[0]
      captured_port = host_port_tuple[1]
      captured_timeout_sec = timeout_sec
      return fake_socket
    monkeypatch.setattr(socket, 'create_connection', capture_create_connection)
    endpoint = TcpEndpoint(expected_host, expected_port, expected_timeout)
    sut = TcpDevice(endpoint)

    sut.process(Instruction(Command.QUERY_PANEL_LOCK))

    assert captured_host == expected_host
    assert captured_port == expected_port
    assert captured_timeout_sec == expected_timeout

  def test_process_dispatches_single_instruction(self, monkeypatch: pytest.MonkeyPatch):
    instruction = Instruction(Command.QUERY_PANEL_LOCK)
    expected_bytes = Codec.encode(instruction)
    fake_socket = self.stub_socket(monkeypatch)
    sut = self.create_device()

    sut.process(instruction)

    assert fake_socket.request_bytes[0] == expected_bytes

  def test_process_dispatches_multiple_instruction(self, monkeypatch: pytest.MonkeyPatch):
    instruction1 = Instruction(Command.QUERY_PANEL_LOCK)
    expected_bytes1 = Codec.encode(instruction1)
    instruction2 = Instruction(Command.QUERY_OUTPUT_STATUS)
    expected_bytes2 = Codec.encode(instruction2)
    instructions = [instruction1, instruction2]
    fake_socket = self.stub_socket(monkeypatch)
    sut = self.create_device()

    sut.process(instructions)

    assert fake_socket.send_count == 2
    assert fake_socket.request_bytes[0] == expected_bytes1
    assert fake_socket.request_bytes[1] == expected_bytes2

  def test_process_continues_without_erroring_when_no_response_received(
        self,
        caplog: pytest.LogCaptureFixture,
        monkeypatch: pytest.MonkeyPatch
      ):
      caplog.set_level(logging.INFO)
      instruction1 = Instruction(Command.QUERY_PANEL_LOCK)
      instruction2 = Instruction(Command.QUERY_OUTPUT_STATUS)
      instructions = [instruction1, instruction2]
      fake_socket = self.stub_socket(monkeypatch)
      fake_socket.should_timeout = True
      sut = self.create_device()

      result = sut.process(instructions)

      assert 'Timed out' in caplog.text
      assert fake_socket.send_count == 2
      assert len(result) == 0
    
  def test_process_returns_multiple_response_instructions(self, monkeypatch: pytest.MonkeyPatch):
    response_bytes = b'\x5e\x80\x80\x81\x5f\x80\x80\x81'
    instruction = Instruction(Command.QUERY_PANEL_LOCK)
    expected_results = [
      Instruction(Command.PANEL_LOCK, 0, 0, 1),
      Instruction(Command.QUERY_PANEL_LOCK, 0, 0, 1)
    ]
    fake_socket = self.stub_socket(monkeypatch)
    fake_socket.response_bytes = response_bytes
    sut = self.create_device()

    results = sut.process(instruction)

    assert results == expected_results

  def test_process_closes_connection_when_complete(self, monkeypatch: pytest.MonkeyPatch):
    instruction = Instruction(Command.QUERY_OUTPUT_STATUS)
    fake_socket = self.stub_socket(monkeypatch)
    sut = self.create_device()

    sut.process(instruction)

    assert fake_socket.was_closed

  def stub_socket(self, patch: pytest.MonkeyPatch) -> FakeSocket:
    fake_socket = FakeSocket()
    patch.setattr(socket, 'create_connection', lambda *_: fake_socket)
    return fake_socket

  def create_device(self, endpoint: TcpEndpoint = TcpEndpoint('localhost')):
    return TcpDevice(endpoint)

#
# Helpers
#
def gen_invalid_cmd_id(rand = random) -> int:
  return rand.randrange(64, 128)

def gen_valid_cmd(rand = random) -> Command:
  return random.choice(list(Command))

def gen_valid_io_value(rand = random) -> int:
  return rand.choice(_VALID_RANGE)

def gen_invalid_io_value(rand = random) -> int:
  max_valid = _VALID_RANGE.stop
  return rand.randrange(max_valid, max_valid + 100)

def gen_port(rand = random) -> int:
  return rand.randrange(49152, 65535)




from typing import Protocol

class MediaSwitch(Protocol):
  """
  Representation of a multi-input, single- or fixed-output media switch, such as a
  5x1 HDMI switch, or a 5x2 HDMI switch where the output on both ports is always the same.

  Example Kramer devices: VS-41H, VS-81H, VS-161H
  """

  def select_source(self, input: int) -> None:
    """
    Select the specified video input
    """

  def lock(self):
     """
     Lock front panel
     """

  def unlock(self):
     """
     Unlock front panel
     """

  def update(self) -> None:
    """
    Refresh device state.
    """

  @property
  def selected_source(self) -> int:
    """
    Returns the input number of the selected source
    """

  @property
  def is_locked(self) -> bool:
    """ Returns `true` when front panel is locked, `false` otherwise.
    
    Note that the device can still be controlled remotely when front panel is
    locked.
    """

  @property
  def input_count(self) -> int:
    """
    Returns the number of inputs the switch has
    """

  @property
  def output_count(self) -> int:
    """
    Returns the number of outputs the switch has
    """



from typing import Optional

from kesslerav.protocol2k.io import Command, Instruction
from kesslerav.protocol2k.media_switch import MediaSwitch

from fakes import FakeDevice

class TestMediaSwitch:
  def test_sets_machine_id_when_specified(self):
    expected_machine_id = 2

    (sut, _) = self.create_media_switch(machine_id = expected_machine_id)

    assert sut.machine_id == expected_machine_id

  def test_initializes_state_from_device(self):
    input_count = 8
    output_count = 2
    selected_source = 3
    fake_device = FakeDevice()
    response_instructions = [
      Instruction(Command.DEFINE_MACHINE, 1, input_count, 1),
      Instruction(Command.DEFINE_MACHINE, 2, output_count, 1),
      Instruction(Command.QUERY_OUTPUT_STATUS, 0, selected_source, 1),
      Instruction(Command.QUERY_PANEL_LOCK, 0, 1, 1),
    ]
    fake_device.response_instructions = response_instructions

    (sut, _) = self.create_media_switch(device = fake_device)

    assert sut.input_count == input_count
    assert sut.output_count == output_count
    assert sut.selected_source == selected_source
    assert sut.is_locked

  def test_update_reloads_state_from_device(self):
    input_count = 8
    output_count = 2
    selected_source = 3
    fake_device = FakeDevice()
    response_instructions = [
      Instruction(Command.DEFINE_MACHINE, 1, input_count, 1),
      Instruction(Command.DEFINE_MACHINE, 2, output_count, 1),
      Instruction(Command.QUERY_OUTPUT_STATUS, 0, selected_source, 1),
      Instruction(Command.QUERY_PANEL_LOCK, 0, 1, 1),
    ]
    fake_device.response_instructions = response_instructions

    (sut, _) = self.create_media_switch(device = fake_device)

    assert sut.input_count == input_count
    assert sut.output_count == output_count
    assert sut.selected_source == selected_source
    assert sut.is_locked

    # Device state changed
    input_count2 = 16
    output_count2 = 1
    selected_source2 = 12
    response_instructions2 = [
      Instruction(Command.DEFINE_MACHINE, 1, input_count2, 1),
      Instruction(Command.DEFINE_MACHINE, 2, output_count2, 1),
      Instruction(Command.QUERY_OUTPUT_STATUS, 0, selected_source2, 1),
      Instruction(Command.QUERY_PANEL_LOCK, 0, 0, 1),
    ]
    fake_device.response_instructions = response_instructions2

    sut.update()

    assert sut.input_count == input_count2
    assert sut.output_count == output_count2
    assert sut.selected_source == selected_source2
    assert not sut.is_locked

  def test_selected_source_sends_switch_video_instruction(self):
    input_count = 8
    output_count = 2
    initial_source = 7
    selected_source = 3
    fake_device = FakeDevice()
    response_instructions = [
      Instruction(Command.DEFINE_MACHINE, 1, input_count, 1),
      Instruction(Command.DEFINE_MACHINE, 2, output_count, 1),
      Instruction(Command.QUERY_OUTPUT_STATUS, 0, initial_source, 1),
    ]
    expected = [Instruction(Command.SWITCH_VIDEO, selected_source, 0, 1)]
    fake_device.response_instructions = response_instructions

    (sut, _) = self.create_media_switch(device = fake_device)
    fake_device.clear_instructions()

    sut.select_source(selected_source)

    assert sut.selected_source == selected_source
    assert fake_device.processed_instructions == expected

  def test_selected_source_normalizes_negative_value_to_0(self):
    input_count = 8
    output_count = 2
    initial_source = 7
    selected_source = -5
    expected_source = 0
    fake_device = FakeDevice()
    response_instructions = [
      Instruction(Command.DEFINE_MACHINE, 1, input_count, 1),
      Instruction(Command.DEFINE_MACHINE, 2, output_count, 1),
      Instruction(Command.QUERY_OUTPUT_STATUS, 0, initial_source, 1),
    ]
    expected = [Instruction(Command.SWITCH_VIDEO, expected_source, 0, 1)]
    fake_device.response_instructions = response_instructions

    (sut, _) = self.create_media_switch(device = fake_device)
    fake_device.clear_instructions()

    sut.select_source(selected_source)

    assert sut.selected_source == expected_source
    assert fake_device.processed_instructions == expected

  def test_selected_source_normalizes_sources_above_max_to_last_input(self):
    input_count = 8
    output_count = 1
    initial_source = 3
    selected_source = 10
    expected_source = 8
    fake_device = FakeDevice()
    response_instructions = [
      Instruction(Command.DEFINE_MACHINE, 1, input_count, 1),
      Instruction(Command.DEFINE_MACHINE, 2, output_count, 1),
      Instruction(Command.QUERY_OUTPUT_STATUS, 0, initial_source, 1),
    ]
    expected = [Instruction(Command.SWITCH_VIDEO, expected_source, 0, 1)]
    fake_device.response_instructions = response_instructions

    (sut, _) = self.create_media_switch(device = fake_device)
    fake_device.clear_instructions()

    sut.select_source(selected_source)

    assert sut.selected_source == expected_source
    assert fake_device.processed_instructions == expected

  def test_selected_source_updates_state_from_device(self):
    input_count = 8
    output_count = 1
    initial_source = 3
    selected_source = 6
    expected_source = 8
    fake_device = FakeDevice()
    fake_device.response_instructions = [
      Instruction(Command.DEFINE_MACHINE, 1, input_count, 1),
      Instruction(Command.DEFINE_MACHINE, 2, output_count, 1),
      Instruction(Command.QUERY_OUTPUT_STATUS, 0, initial_source, 1),
      Instruction(Command.QUERY_PANEL_LOCK, 0, 0, 1),
    ]
    expected = [Instruction(Command.SWITCH_VIDEO, selected_source, 0, 1)]
    (sut, _) = self.create_media_switch(device = fake_device)
    fake_device.clear_instructions()
    # Simulates the case where a source was requested, but something else
    # requested a different source (e.g., different thread, front panel) so
    # selected source ended up being different.
    fake_device.response_instructions = [
      Instruction(Command.SWITCH_VIDEO, selected_source, 0, 1),
      Instruction(Command.SWITCH_VIDEO, expected_source, 0, 1),
      Instruction(Command.PANEL_LOCK, 1, 0, 1),
    ]

    sut.select_source(selected_source)

    assert sut.selected_source == expected_source
    assert sut.is_locked
    assert fake_device.processed_instructions == expected

  def test_lock_sends_lock_instruction(self):
    expected = [Instruction(Command.PANEL_LOCK, 1, 0, 1)]
    (sut, fake_device) = self.create_media_switch()
    fake_device.clear_instructions()

    sut.lock()

    assert sut.is_locked
    assert fake_device.processed_instructions == expected

  def test_lock_updates_state_from_device(self):
    input_count = 8
    output_count = 1
    initial_source = 3
    expected_source = 6
    fake_device = FakeDevice()
    fake_device.response_instructions = [
      Instruction(Command.DEFINE_MACHINE, 1, input_count, 1),
      Instruction(Command.DEFINE_MACHINE, 2, output_count, 1),
      Instruction(Command.QUERY_OUTPUT_STATUS, 0, initial_source, 1),
      Instruction(Command.QUERY_PANEL_LOCK, 0, 0, 1),
    ]
    expected = [Instruction(Command.PANEL_LOCK, 1, 0, 1)] # Lock
    (sut, _) = self.create_media_switch(device = fake_device)
    fake_device.clear_instructions()
    # Simulates the case where panel lock was requested, but something else
    # requested a other changes (e.g., different thread, front panel) so
    # panel lock and source end up being different.
    fake_device.response_instructions = [
      Instruction(Command.PANEL_LOCK, 1, 0, 1), # Lock
      Instruction(Command.SWITCH_VIDEO, expected_source, 0, 1),
      Instruction(Command.PANEL_LOCK, 0, 0, 1), # Unlock
    ]

    sut.lock()

    assert sut.selected_source == expected_source
    assert not sut.is_locked
    assert fake_device.processed_instructions == expected

  def test_unlock_sends_unlock_instruction(self):
    expected = [Instruction(Command.PANEL_LOCK, 0, 0, 1)]
    (sut, fake_device) = self.create_media_switch()
    fake_device.clear_instructions()

    sut.unlock()

    assert not sut.is_locked
    assert fake_device.processed_instructions == expected

  def test_unlock_updates_state_from_device(self):
    input_count = 8
    output_count = 1
    initial_source = 6
    expected_source = 2
    fake_device = FakeDevice()
    fake_device.response_instructions = [
      Instruction(Command.DEFINE_MACHINE, 1, input_count, 1),
      Instruction(Command.DEFINE_MACHINE, 2, output_count, 1),
      Instruction(Command.QUERY_OUTPUT_STATUS, 0, initial_source, 1),
      Instruction(Command.QUERY_PANEL_LOCK, 0, 0, 1),
    ]
    expected = [Instruction(Command.PANEL_LOCK, 0, 0, 1)] # Unlock
    (sut, _) = self.create_media_switch(device = fake_device)
    fake_device.clear_instructions()
    # Simulates the case where panel lock was requested, but something else
    # requested a other changes (e.g., different thread, front panel) so
    # panel lock and source end up being different.
    fake_device.response_instructions = [
      Instruction(Command.PANEL_LOCK, 0, 0, 1), # Unlock
      Instruction(Command.SWITCH_VIDEO, expected_source, 0, 1),
      Instruction(Command.PANEL_LOCK, 1, 0, 1), # Lock
    ]

    sut.unlock()

    assert sut.selected_source == expected_source
    assert sut.is_locked
    assert fake_device.processed_instructions == expected
  
  def create_media_switch(
      self,
      device = FakeDevice(),
      machine_id: Optional[int] = 1
    ) -> tuple[MediaSwitch, FakeDevice]:
    return (MediaSwitch(device, machine_id), device)




"""Kramer A/V Protcol 2000 control library"""
from typing import Optional

from .constants import PROTOCOL_2K, SCHEME_TCP
from .url_parser import parse_url
from .media_switch import MediaSwitch

from .protocol2k import get_tcp_media_switch

def get_media_switch(
    url: str,
    timeout_sec: Optional[float] = None,
    machine_id: Optional[int] = None
  ) -> MediaSwitch:
  """
  Create media switch representation whose state can be accessed via the specified
  URL.

  Currently supported:
    + Schemes:
      + TCP (identifier: tcp)
    + Protocols:
      + Protocol 2000 (identifier: protocol2k)

  Defaults:
    + Default scheme: TCP (identifier: tcp)
    + Default TCP port: 5000
    + Default protocol: Protocol 2000 (identifier: protocol2k)

  URL format:
    The URL takes the form of `<scheme>://<host>:<port>#<protocol>` with all
    but the host being optional.

    Examples:
      + 10.0.0.1 ->
          Scheme: tcp, Host: 10.0.0.1, Port: 5000, Protocol: protocol2k
      + localhost:1337 ->
          Scheme: tcp, Host: localhost, Port: 1337, Protocol: protocol2k
      + localhost:1337#protocol2000 ->
          Scheme: tcp, Host: localhost, Port: 1337, Protocol: protocol2k
      + tcp://10.0.0.1 ->
          Scheme: tcp, Host: 10.0.0.1, Port: 5000, Protocol: protocol2k
      + tcp://switch.local:8080 ->
          Scheme: tcp, Host: switch.local, Port: 8080, Protocol: protocol2k
      + tcp://localhost:8080#protocol2k
          Scheme: tcp, Host: localhost, Port: 8080, Protocol: protocol2k

  Args:
    url (str): The URL at which the device state can be accessed.
    timeout_sec (Optional[float]): Timeout, in seconds, to use when communicating
      with the device. Default: None, which allows the underlying protocol driver
      to determine.
    machine_id (Optional[int]): Specifies the machine ID of the target device.
      This is only applicable with serial communications. Default: None, which
      allows the underlying protocol driver to determine.

  Returns:
    MediaSwitch: Representation of the media switch device, including methods for
      controlling it (e.g., selecting sources.)
  """
  endpoint = parse_url(url)
  if endpoint.protocol == PROTOCOL_2K and endpoint.scheme == SCHEME_TCP:
    return get_tcp_media_switch(
      host = endpoint.host,
      port = endpoint.port,
      timeout_sec = timeout_sec,
      machine_id = machine_id
    )
  else:
    raise ValueError(f'Unsupported url specified: {url}')


from kesslerav.constants import PROTOCOL_2K
from kesslerav.url_parser import Endpoint, parse_url, \
  _DEFAULT_HOST, _DEFAULT_PORT_TCP, _DEFAULT_PROTOCOL, \
  _DEFAULT_SCHEME, _PROTOCOL_2K_ALT

class TestEndpoint:
  # Scheme
  def test_sets_tcp_as_default_scheme(self):
    sut = Endpoint()

    assert sut.scheme == _DEFAULT_SCHEME

  def test_sets_scheme_as_specified(self):
    scheme = 'udp'

    sut = Endpoint(scheme = scheme)

    assert sut.scheme == scheme

  # Host
  def test_sets_host_default_when_not_specified(self):
    sut = Endpoint()

    assert sut.host == _DEFAULT_HOST

  def test_sets_host_as_specified(self):
    host = 'localhost'

    sut = Endpoint(host = host)

    assert sut.host == host

  # Port
  def test_sets_port_default_when_not_specified_and_tcp_scheme(self):
    sut = Endpoint()

    assert sut.port == _DEFAULT_PORT_TCP

  def test_leaves_port_unspecified_for_unknown_schemes(self):
    sut = Endpoint(scheme = 'udp')

    assert sut.port is None

  def test_sets_port_as_specified(self):
    port = 1337

    sut = Endpoint(port = port)

    assert sut.port == port

  # Protocol
  def test_sets_protocol_default_when_not_specified(self):
    sut = Endpoint()

    assert sut.protocol == _DEFAULT_PROTOCOL

  def test_sets_protocol_when_specified(self):
    protocol = 'http'

    sut = Endpoint(protocol = protocol)

    assert sut.protocol == protocol

  def test_normalizes_protocol2k_when_alternate_id_used(self):
    protocol = _PROTOCOL_2K_ALT

    sut = Endpoint(protocol = protocol)

    assert sut.protocol == PROTOCOL_2K

class TestParseUrl:
  def test_host_only_parses_correctly(self):
    url = 'localhost'

    result = parse_url(url)

    assert result.scheme == _DEFAULT_SCHEME
    assert result.host == url
    assert result.port == _DEFAULT_PORT_TCP
    assert result.protocol == _DEFAULT_PROTOCOL

  def test_host_and_port_only_parses_correctly(self):
    host = '10.0.0.1'
    port = 1337
    url = f'{host}:{port}'

    result = parse_url(url)

    assert result.scheme == _DEFAULT_SCHEME
    assert result.host == host
    assert result.port == port
    assert result.protocol == _DEFAULT_PROTOCOL

  def test_scheme_and_host_parses_correctly(self):
    scheme = 'udp'
    host = '192.168.0.1'
    url = f'{scheme}://{host}'

    result = parse_url(url)

    assert result.scheme == scheme
    assert result.host == host
    assert result.port is None
    assert result.protocol == _DEFAULT_PROTOCOL

  def test_scheme_host_and_port_parses_correctly(self):
    scheme = 'udp'
    host = '192.168.0.1'
    port = 50000
    url = f'{scheme}://{host}:{port}'

    result = parse_url(url)

    assert result.scheme == scheme
    assert result.host == host
    assert result.port == port
    assert result.protocol == _DEFAULT_PROTOCOL

  def test_scheme_host_port_and_protocol_parses_correctly(self):
    scheme = 'http'
    host = 'mediaswitch.local'
    port = 1337
    protocol = 'l33tsp34k'
    url = f'{scheme}://{host}:{port}#{protocol}'

    result = parse_url(url)

    assert result.scheme == scheme
    assert result.host == host
    assert result.port == port
    assert result.protocol == protocol

  def test_host_and_protocol_parses_correctly(self):
    host = 'mediaswitch.local'
    protocol = 'l33tsp34k'
    url = f'{host}#{protocol}'

    result = parse_url(url)

    assert result.scheme == _DEFAULT_SCHEME
    assert result.host == host
    assert result.port == _DEFAULT_PORT_TCP
    assert result.protocol == protocol

  def test_host_port_and_protocol_parses_correctly(self):
    host = 'mediaswitch.local'
    protocol = 'l33tsp34k'
    port = 1337
    url = f'{host}:{port}#{protocol}'

    result = parse_url(url)

    assert result.scheme == _DEFAULT_SCHEME
    assert result.host == host
    assert result.port == port
    assert result.protocol == protocol



from logging import getLogger as get_logger
from typing import Optional

from ..media_switch import MediaSwitch as MediaSwitchProtocol
from .io import Command, Instruction, TcpDevice

_LOGGER = get_logger(__name__)

class MediaSwitch(MediaSwitchProtocol):
  def __init__(self, device: TcpDevice, machine_id: Optional[int] = None):
    self._device = device 
    self._machine_id = machine_id
    self._is_locked = False
    self._selected_source = 0
    self._input_count = 0
    self._output_count = 0
    self.update()

  def select_source(self, input: int) -> None:
    """
    Select the specified video input
    """
    normalized_input = input
    if normalized_input < 0:
      normalized_input = 0
    elif normalized_input > self._input_count:
      normalized_input = self._input_count

    instruction = Instruction(Command.SWITCH_VIDEO, normalized_input, None, self._machine_id)
    self._selected_source = normalized_input
    self._process(instruction)

  def lock(self):
    """
    Lock panel
    """
    instruction = Instruction(Command.PANEL_LOCK, 1, None, self._machine_id)
    self._is_locked = True
    self._process(instruction)

  def unlock(self):
    """
    Unlock panel
    """
    instruction = Instruction(Command.PANEL_LOCK, 0, None, self._machine_id)
    self._is_locked = False
    self._process(instruction)
  
  def update(self) -> None:
    self._process(self._update_instructions())

  @property
  def selected_source(self) -> int:
    """
    Returns the input number of the selected source
    """
    return self._selected_source

  @property
  def input_count(self) -> int:
    """
    The number of inputs the switch has
    """
    return self._input_count

  @property
  def output_count(self) -> int:
    """
    The number of outputs the switch has
    """
    return self._output_count

  @property
  def is_locked(self) -> bool:
    """
    Returns `true` when panel is locked, `false` otherwise.
    """
    return self._is_locked

  @property
  def machine_id(self) -> int | None :
    return self._machine_id

  def _process(self, instructions: list[Instruction] | Instruction) -> None:
    results = self._device.process(instructions)
    self._update_from_instructions(results)

  def _update_from_instructions(self, instructions: list[Instruction]) -> None:
    for instruction in instructions:
      match instruction.id:
        case Command.DEFINE_MACHINE:
          if instruction.input_value == 1:
            self._input_count = instruction.output_value
          elif instruction.input_value == 2:
            self._output_count = instruction.output_value
        case Command.PANEL_LOCK:
          self._is_locked = (instruction.input_value == 1)
        case Command.SWITCH_VIDEO:
          self._selected_source = instruction.input_value
        case Command.QUERY_OUTPUT_STATUS:
          self._selected_source = instruction.output_value
        case Command.QUERY_PANEL_LOCK:
          self._is_locked = (instruction.output_value == 1)
        case _:
          _LOGGER.info(f'Discarded instruction: {instruction}')
  
  def _update_instructions(self) -> list[Instruction]:
    return [
      # Queries the number of inputs
      Instruction(Command.DEFINE_MACHINE, 1, 1, self._machine_id),
      # Queries the number of outputs
      Instruction(Command.DEFINE_MACHINE, 2, 1, self._machine_id),
      # Queries which input is currently being routed to output 1
      Instruction(Command.QUERY_OUTPUT_STATUS, 0, 1, self._machine_id),
      # Queries the panel lock status
      Instruction(Command.QUERY_PANEL_LOCK, None, None, self._machine_id),
    ]





import itertools
import socket
import struct

from enum import IntEnum, unique
from logging import getLogger as get_logger
from typing import Optional

_LOGGER = get_logger(__name__)

@unique
class Command(IntEnum):
  """
  Enumerates the supported Protocol 2000 commands
  """
  SWITCH_VIDEO = 1
  RECALL_VIDEO_STATUS = 4
  ERROR = 16
  PANEL_LOCK = 30
  QUERY_OUTPUT_STATUS = 5
  QUERY_PANEL_LOCK = 31
  IDENTIFY_MACHINE = 61
  DEFINE_MACHINE = 62

  @classmethod
  def is_supported(cls, cmd_id: int) -> bool:
    return cmd_id in iter(Command)

# Validation rules: limit I/O values to one byte
_VALUE_MIN = 0
_VALUE_MAX = 128 # Only 7 bits available for data transport
_VALID_RANGE: range = range(_VALUE_MIN, _VALUE_MAX)
  
def _validated_value(maybe_value: Optional[int], default_value: int = 0) -> int:
  if maybe_value is None:
    return default_value
  if maybe_value not in _VALID_RANGE:
    raise ValueError(
      f'Valid values are between {_VALUE_MIN} and {_VALUE_MAX - 1}, '
      f'inclusive. Received: {maybe_value}'
    )
  else:
    return maybe_value

class Instruction:
  """
  Encapsulates a fully-formed Protocol 2000 instruction
  """

  # Defaults to the override value, meaning ALL machines receiving the instruction
  # will respond, regardless of machine ID setting.
  DEFAULT_MACHINE_ID: int = 0b01000001

  # Default name for unsupported/unrecognized commands
  UNSUPPORTED_COMMAND_NAME: str = "UNSUPPORTED"

  # Protocol 2000 uses 4-byte instructions
  SIZE_BYTES: int = 4
  FORMAT: str = '!I' # 4-bytes

  def __init__(
    self,
    cmd: int,
    input_value: Optional[int] = None,
    output_value: Optional[int] = None,
    maybe_machine_id: Optional[int] = None
  ):
    if Command.is_supported(cmd):
      self._command = Command(cmd)
      self._unsupported_command_id = None
    else:
      self._command = None
      self._unsupported_command_id = cmd

    self._input_value = _validated_value(input_value)
    self._output_value = _validated_value(output_value)
    self._machine_id = _validated_value(
      maybe_machine_id,
      Instruction.DEFAULT_MACHINE_ID
    )
    
  @property
  def id(self) -> int:
    if self._command is not None:
      return self._command.value
    else:
      return self._unsupported_command_id

  @property
  def name(self) -> str:
    if self._command is not None:
      return self._command.name
    else:
      return Instruction.UNSUPPORTED_COMMAND_NAME

  @property
  def input_value(self) -> int:
    return self._input_value

  @property
  def output_value(self) -> int:
    return self._output_value

  @property
  def machine_id(self) -> int:
    return self._machine_id

  @property
  def is_supported(self) -> bool:
    return self._command is not None

  @property
  def frame(self) -> list[int]:
    return [
      self.id,
      self.input_value,
      self.output_value,
      self.machine_id
    ]

  def __str__(self) -> str:
    return (
      f'<Instruction id: {self.id} name: {self.name} input: {self.input_value} '
      f'output: {self.output_value} machine_id: {self.machine_id}>'
    )

  def __repr__(self) -> str:
    return (
      f'Instruction<{self.name}>({self.id}, {self.input_value}, '
      f'{self.output_value}, {self.machine_id})'
    )

  def __eq__(self, other):
    if not isinstance(other, Instruction):
      return NotImplemented
    return self.frame == other.frame

class Codec:
  """
  Bidirectionally converts between Instruction and bytes
  """

  @classmethod
  def encode(cls, instruction: Instruction) -> bytes:
    msg = cls._encode_message(instruction)
    data = bytes(msg)
    return data
  
  @classmethod
  def decode(cls, data: bytes) -> Instruction:
    frame = cls._decode_message(data)
    cmd = Instruction(*frame)
    return cmd

  @classmethod
  def _encode_message(cls, instruction: Instruction) -> list[int]:
    cmd_id, *values = instruction.frame
    encoded_values = list(map(cls._encode_value, values))
    return [cmd_id] + encoded_values
  
  @classmethod
  def _decode_message(cls, data: bytes) -> list[int]:
    cmd_id, *encoded_values = [byte for byte in data]
    if not Command.is_supported(cmd_id):
      # Command ID is likely a response-encoded ID; decode it
      cmd_id = cls._decode_command_id(cmd_id)
    values = list(map(cls._decode_value, encoded_values))
    return [cmd_id] + values 

  # Per protocol, the first bit for all I/O values must be 1
  @classmethod
  def _encode_value(cls, value: int) -> int:
    return 0b10000000 | value

  # Values must have first bit set to 1, so flipping it back yields the original
  # value.
  @classmethod
  def _decode_value(cls, value: int) -> int:
    return value ^ 0b10000000

  # Response command ID is the request with the second bit set to 1. To decode,
  # flip it back to determine corresponding request command ID.
  @classmethod
  def _decode_command_id(cls, command_id: int) -> int:
    return command_id ^ 0b01000000

class TcpEndpoint:
  """
  Protocol 2000 TCP endpoint location details
  """
  DEFAULT_PORT: int = 5000
  DEFAULT_TIMEOUT_SEC: float = 0.250

  def __init__(
      self,
      host: str,
      port: Optional[int] = None,
      timeout_sec: Optional[float] = DEFAULT_TIMEOUT_SEC
    ):
    self._host = host
    if port is None:
      self._port = TcpEndpoint.DEFAULT_PORT
    else:
      self._port = port
    self._timeout_sec = timeout_sec

  @property
  def host(self) -> str:
    return self._host

  @property
  def port(self) -> int:
    return self._port

  @property
  def timeout_sec(self) -> Optional[float]:
    return self._timeout_sec

class TcpDevice:
  """
  Manages TCP I/O for a specific Protocol 2000-based device
  """
  # Maximum number of response instructions to read
  PAGE_SIZE: int = 32
  # Size of the buffer used to read responses
  BUFFER_SIZE_BYTES: int = Instruction.SIZE_BYTES * PAGE_SIZE

  def __init__(
      self,
      endpoint: TcpEndpoint,
    ):
    self._endpoint = endpoint

  def process(self, instructions: list[Instruction] | Instruction) -> list[Instruction]:
    try:
      _ = iter(instructions)
    except TypeError:
      # Single instruction provided; wrap it.
      instructions = [instructions]
    results = []

    conn = self._create_connection()
    try:
      for instruction in instructions:
        result = self._execute_instruction(instruction, conn)
        results.append(result)
    except Exception as ex:
      _LOGGER.error(f'Failed communicating with device: {ex}')
    finally:
      conn.close()

    # Results is a list of lists, so flatten before returning
    flat_results = list(itertools.chain.from_iterable(results))
    return flat_results
  
  def _create_connection(self) -> socket.socket:
    return socket.create_connection(
      (self._endpoint.host, self._endpoint.port),
      self._endpoint.timeout_sec
    )
  
  def _execute_instruction(
      self,
      instruction: Instruction,
      conn: socket.socket
    ) -> list[Instruction]:
    req_bytes = Codec.encode(instruction)
    conn.send(req_bytes)

    # Device can return multiple instructions when its physical controls are
    # used. To capture them all (to reconstruct device state) we read using a
    # buffer that can hold multiple instructions and then return them all in
    # chronological event order.
    result: list[Instruction] = []
    try:
      while len(result) < 1:
        data = conn.recv(TcpDevice.BUFFER_SIZE_BYTES)
        responses = struct.iter_unpack(Instruction.FORMAT, data)
        for response in responses:
          resp_bytes = response[0].to_bytes(Instruction.SIZE_BYTES)
          instruction = Codec.decode(resp_bytes)
          result.append(instruction)
    except TimeoutError:
      _LOGGER.info(
        'Timed out waiting for response. Ignoring, since another thread may '
        'have processed the response already.'
      )

    return result



"""
Encapsulates Protocol 2000 communication and devices
"""
from typing import Optional

from ..media_switch import MediaSwitch as MediaSwitchProtocol
from .io import TcpDevice, TcpEndpoint
from .media_switch import MediaSwitch

def get_tcp_media_switch(
    host: str,
    port: Optional[int] = None,
    timeout_sec: Optional[float] = None,
    machine_id: Optional[int] = None
  ) -> MediaSwitchProtocol:
  if timeout_sec is None:
    # Let `TcpEndpoint` manage timeout
    endpoint = TcpEndpoint(host, port)
  else:
    endpoint = TcpEndpoint(host, port, timeout_sec)
  tcp_device = TcpDevice(endpoint)
  return MediaSwitch(tcp_device, machine_id)


from typing import Optional
from urllib.parse import urlparse

from .constants import PROTOCOL_2K, SCHEME_TCP

_PROTOCOL_2K_ALT = 'protocol2000'

_DEFAULT_HOST = 'localhost'
_DEFAULT_PORT_TCP = 5000
_DEFAULT_PROTOCOL = PROTOCOL_2K
_DEFAULT_SCHEME = SCHEME_TCP

def _is_empty(value: str) -> bool:
  return (value is None or len(value) == 0)


class Endpoint:
  def __init__(
      self,
      scheme: Optional[str] = None,
      host: Optional[str] = None,
      port:Optional[int] = None,
      protocol: Optional[str] = None
    ):
    if _is_empty(scheme):
      self._scheme = _DEFAULT_SCHEME
    else:
      self._scheme = scheme

    if _is_empty(host):
      self._host = _DEFAULT_HOST
    else:
      self._host = host

    if port is None:
      self._port = self._default_port(self._scheme)
    else:
      self._port = port

    if _is_empty(protocol) or protocol == _PROTOCOL_2K_ALT:
      self._protocol = _DEFAULT_PROTOCOL
    else:
      self._protocol = protocol

  @property
  def scheme(self):
    return self._scheme

  @property
  def host(self):
    return self._host

  @property
  def port(self):
    return self._port

  @property
  def protocol(self):
    return self._protocol

  def _default_port(self, scheme: str) -> Optional[int]:
    if scheme == SCHEME_TCP:
      return _DEFAULT_PORT_TCP
    return None


def parse_url(url: str) -> Endpoint:
  normalized_url = url
  if '://' not in url:
    normalized_url = f'{_DEFAULT_SCHEME}://{url}'
  parsed_url = urlparse(normalized_url)
  endpoint = Endpoint(
    scheme = parsed_url.scheme,
    host = parsed_url.hostname,
    port = parsed_url.port,
    protocol = parsed_url.fragment
  )
  return endpoint



from kesslerav.protocol2k.io import Instruction

#
# Fakes
#
class FakeSocket:
  FAKE_INSTRUCTION_BYTES = b'\x50\x80\x80\x81'

  def __init__(self):
    self.request_bytes: list[bytes] = []
    self.response_bytes: bytes = FakeSocket.FAKE_INSTRUCTION_BYTES
    self.response_buffer_size = 0
    self.response_index = 0

    self.should_timeout = False

    self.send_count = 0
    self.recv_count = 0
    self.close_count= 0
    self.was_closed = False

  def send(self, data: bytes) -> None:
    self.send_count += 1
    self.request_bytes.append(data)

  def recv(self, bufsize: int) -> bytes:
    self.recv_count += 1
    self.response_buffer_size = bufsize
    if self.should_timeout:
      raise TimeoutError
    else:
      return self.response_bytes

  def close(self) -> None:
    self.close_count += 1
    self.was_closed = True

class FakeDevice:
  def __init__(self):
    self.processed_instructions = []
    self.response_instructions = []

    self.process_count = 0

  def process(self, instructions: list[Instruction] | Instruction) -> list[Instruction]:
    try:
      _ = iter(instructions)
    except TypeError:
      instructions = [instructions]
    self.process_count += 1
    self.processed_instructions.extend(instructions)
    return self.response_instructions

  def clear_processed_instructions(self):
    self.processed_instructions = []

  def clear_response_instructions(self):
    self.response_instructions = []

  def clear_instructions(self):
    self.clear_processed_instructions()
    self.clear_response_instructions()